#! /bin/sh

# SPDX-FileCopyrightText: none
# SPDX-License-Identifier: CC0-1.0

#This file outputs in a separate line each file with a .desktop syntax
#that needs to be translated but has a non .desktop extension
find -name \*.kdevtemplate -print
