/*
    SPDX-FileCopyrightText: 2021 Alexander Lohnau <alexander.lohnau@gmx.de>
    SPDX-License-Identifier: LGPL-2.0-or-later
*/

#include "fakerunner.h"

K_PLUGIN_CLASS_WITH_JSON(FakeRunner, "metadatafile1.json")

#include "fakerunnerplugin.moc"
